import warnings

import shapely.geometry as sh

from .link_component import Vertices
from .link import Conduit
from .node_component import Coordinate
from .subcatch import Polygon
from ..section_labels import CONDUITS, VERTICES, COORDINATES, POLYGONS
from ..helpers import InpSection
from ..inp import SwmmInput


class InpSectionGeo(InpSection):
    """child class of :obj:`swmm_api.input_file.helpers.InpSection`. See parent class for all functions."""
    def __init__(self, section_object, crs="EPSG:32633"):
        """
        create a section for ``.inp``-file with geo-objects (i.e. nodes, links, subcatchments, raingages, ...)

        Args:
            section_object (BaseSectionObject-like): object class which is stored in this section.
                This information is used to set the index of the section and
                to decide if the section can be exported (converted to a string) as a table.
            crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.
        """
        InpSection.__init__(self, section_object)
        self._crs = crs

    def set_crs(self, crs):
        """
        Set the Coordinate Reference System (CRS) of a geo-section.

        Notes:
            The underlying geometries are not transformed to this CRS.

        Args:
            crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.
        """
        self._crs = crs

    @property
    def geo_series(self):
        """
        Get a geopandas.GeoSeries representation for the geo-section.
        This function sets the object default crs.

        Returns:
            geopandas.GeoSeries: geo-series of the section-data
        """
        return self.get_geo_series(self._crs)

    def get_geo_series(self, crs):
        """
        get a geopandas.GeoSeries representation for the geo-section using a custom crs.

        Args:
            crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.

        Returns:
            geopandas.GeoSeries: geo-series of the section-data
        """
        from geopandas import GeoSeries
        return GeoSeries({label: item.geo for label, item in self.items()}, crs=crs, name='geometry')


########################################################################################################################
class CoordinateGeo(Coordinate):
    """child class of :obj:`.node_component.Coordinate`. See parent class for all functions."""
    _section_class = InpSectionGeo

    @property
    def geo(self):
        """
        get the shapely representation for the object (Point).

        Returns:
            shapely.geometry.Point: point object for the coordinates.
        """
        return sh.Point(self.point)

    @classmethod
    def create_section_from_geoseries(cls, data):
        """
        create a COORDINATES inp-file section for a geopandas.GeoSeries

        Args:
            data (geopandas.GeoSeries): geopandas.GeoSeries of coordinates

        Returns:
            InpSectionGeo: COORDINATES inp-file section
        """
        return cls.create_section(zip(data.index, data.x, data.y))

    @classmethod
    def from_shapely(cls, Node, point: sh.Point):
        return cls(Node, point.x, point.y)


class VerticesGeo(Vertices):
    """child class of :obj:`.link_component.Vertices`. See parent class for all functions."""
    _section_class = InpSectionGeo

    @property
    def geo(self):
        """
        get the shapely representation for the object (LineString).

        Returns:
            shapely.geometry.LineString: LineString object for the vertices.
        """
        return sh.LineString(self.vertices)

    @classmethod
    def create_section_from_geoseries(cls, data) :
        """
        create a VERTICES inp-file section for a geopandas.GeoSeries

        Args:
            data (geopandas.GeoSeries): geopandas.GeoSeries of coordinates

        Returns:
            InpSectionGeo: VERTICES inp-file section
        """
        # geometry mit MultiLineString deswegen v[0] mit ersten und einzigen linestring zu verwenden
        s = cls.create_section()
        # s.update({i: Vertices(i, v) for i, v in zip(data.index, map(lambda i: list(i.coords), data.values))})
        # s.add_multiple(cls(i, list(v.coords)) for i, v in data.to_dict().items())
        s.add_multiple(cls.from_shapely(i, v) for i, v in data.items())
        return s

    @classmethod
    def from_shapely(cls, Link, line: sh.LineString):
        return cls(Link, list(line.coords))


class PolygonGeo(Polygon):
    """child class of :obj:`.subcatch.Polygon`. See parent class for all functions."""
    _section_class = InpSectionGeo

    @property
    def geo(self):
        """
        get the shapely representation for the object (Polygon).

        Returns:
            shapely.geometry.Polygon: LineString object for the polygon.
        """
        return sh.Polygon(self.polygon)

    @classmethod
    def create_section_from_geoseries(cls, data):
        """
        create a POLYGONS inp-file section for a geopandas.GeoSeries

        Warnings:
            Only uses the exterior coordinates and ignoring all interiors.

        Args:
            data (geopandas.GeoSeries): geopandas.GeoSeries of polygons

        Returns:
            InpSectionGeo: POLYGONS inp-file section
        """
        s = cls.create_section()
        has_interiors = data.interiors.apply(len) > 0
        if has_interiors.any():
            warnings.warn('Converting GeoSeries with Interiors(Holes) to POLYGON inp-section will ignore this interiors!')
        # s.add_multiple(cls(i, [xy[0:2] for xy in list(p.coords)]) for i, p in data.exterior.iteritems())
        s.add_multiple(cls.from_shapely(i, p) for i, p in data.items())
        return s

    @classmethod
    def from_shapely(cls, Subcatch, polygon: sh.Polygon):
        return cls(Subcatch, cls.convert_shapely(polygon))

    @staticmethod
    def convert_shapely(polygon: sh.Polygon):
        return [xy[0:2] for xy in list(polygon.exterior.coords)]


########################################################################################################################
def convert_section_to_geosection(section, crs="EPSG:32633"):
    """
    converting the section into a geo-section

    Args:
        section (InpSection): section (only `VERTICES`, `COORDINATES`, `POLYGONS` sections)
        crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.

    Returns:
        InpSectionGeo: geo-section
    """
    di = {Coordinate: CoordinateGeo,
          Vertices: VerticesGeo,
          Polygon: PolygonGeo}
    # base object of the section
    old_type = section._section_object  # type: swmm_api.input_file.helpers.BaseSectionObject

    # if already a geo-section
    if old_type in di.values():
        section.set_crs(crs)
        return section

    new_type = di[old_type]

    # create new section and set crs.
    new = new_type.create_section()  # type: InpSectionGeo
    new._data = {k: new_type(**vars(section[k])) for k in section}
    new.set_crs(crs)
    return new


# GeoPandas support for sections
geo_section_converter = {COORDINATES: CoordinateGeo, VERTICES: VerticesGeo, POLYGONS: PolygonGeo}


class SwmmInputGeo(SwmmInput):
    """
    overall class for an input file

    child class of dict

    just used for the copy function and to identify ``.inp``-file data
    
    added geo support
    """""
    @classmethod
    def read_file(cls, filename, custom_converter=None, force_ignore_case=False, encoding='iso-8859-1'):
        return super().read_file(filename, custom_converter=geo_section_converter, force_ignore_case=False, encoding='iso-8859-1')

# Find: SwmmInput\.read_file\(([^,]*)\, custom_converter\=geo_section_converter\)
# Replace: SwmmInputGeo.read_file($1)


def add_geo_support(inp, crs="EPSG:32633"):
    """
    add geo-support to the inp-data

    Warnings:
        only for `VERTICES`, `COORDINATES`, `POLYGONS` sections

    Args:
        inp:
        crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.
    """
    for sec in [VERTICES, COORDINATES, POLYGONS]:
        if (sec in inp) and not isinstance(inp[sec], InpSectionGeo):
            inp[sec] = convert_section_to_geosection(inp[sec], crs=crs)


########################################################################################################################
def section_to_geopandas(section, crs="EPSG:32633"):
    """
    convert a section into a GeoSeries.

    Warnings:
        only for `VERTICES`, `COORDINATES`, `POLYGONS` sections

    Args:
        section (InpSection | InpSectionGeo): section (only `VERTICES`, `COORDINATES`, `POLYGONS`)
        crs: Coordinate Reference System of the geometry objects.
                Can be anything accepted by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
                such as an authority string (eg “EPSG:32633”) or a WKT string.

    Returns:
        geopandas.GeoSeries: geo-data
    """
    return convert_section_to_geosection(section).get_geo_series(crs)


########################################################################################################################
def remove_coordinates_from_vertices(inp):
    # SNIPPET ?!?
    new_vertices_section = dict()
    for link in inp[VERTICES]:  # type: str
        conduit = inp[CONDUITS][link]  # type: Conduit
        new_vertices = list()
        # n1 = inp[COORDINATES][conduit.FromNode]
        new_vertices.append(inp[COORDINATES][conduit.FromNode])
        new_vertices += inp[VERTICES][link].vertices
        new_vertices.append(inp[COORDINATES][conduit.ToNode])
        new_vertices_section[link] = new_vertices
    return new_vertices_section



from .generic_section import (OptionSection, EvaporationSection, TemperatureSection, ReportSection,
                              MapSection, FilesSection, BackdropSection, AdjustmentsSection, TitleSection)

from .link import Conduit, Pump, Orifice, Weir, Outlet
from .link_component import CrossSection, Loss, Vertices

from .node import Junction, Storage, Outfall  # , Divider
from .node_component import DryWeatherFlow, Inflow, Coordinate, RainfallDependentInfiltrationInflow, Treatment

from .others import (RainGage, Control, Transect, Pattern, Pollutant, Symbol, Curve, Timeseries, Tag, Hydrograph,
                     BuildUp, WashOff, LandUse, Label, Aquifer, SnowPack, TimeseriesFile, TimeseriesData)

from .subcatch import (SubArea, SubCatchment, Infiltration, InfiltrationHorton, InfiltrationCurveNumber,
                       InfiltrationGreenAmpt, Polygon, Loading, Coverage, GroundwaterFlow, Groundwater, )

from .lid import LIDControl, LIDUsage



from .collection import nodes_dict, links_dict
from .graph import inp_to_graph, next_nodes, previous_links, next_links
from ..inp import SwmmInput
from ..section_labels import TITLE
from .._type_converter import is_equal
from ..helpers import BaseSectionObject, _sort_by


def compare_sections(s1, s2, precision=3):
    """
    compare two inp file sections and get the differences as string output

    Args:
        s1 (InpSection): filename for the first inp file
        s2 (InpSection): filename for the second inp file
        precision (int): number of relevant decimal places

    Returns:
        str: differences of the sections
    """
    i1 = set(s1.keys())
    i2 = set(s2.keys())
    s_warnings = str()
    not_in_1 = list()
    not_in_2 = list()
    n_equal = 0
    not_equal = list()

    for key in sorted(i1 | i2):
        if (key in i1) and (key in i2):
            if s1[key] == s2[key]:
                n_equal += 1
            else:
                try:
                    if not isinstance(s1[key], BaseSectionObject):
                        not_equal.append(f'{key}: {s1[key]} != {s2[key]}')
                    else:
                        diff = list()
                        for param in s1[key].to_dict_():
                            if not is_equal(s1[key][param], s2[key][param], precision=precision):
                                diff.append(f'{param}=({s1[key][param]} != {s2[key][param]})')
                        if diff:
                            not_equal.append(f'{key}: ' + ' | '.join(diff))
                except:
                    not_equal.append(f'{key}: can not compare')

        # -----------------------------
        elif (key in i1) and (key not in i2):
            not_in_2.append(str(key))
        elif (key not in i1) and (key in i2):
            not_in_1.append(str(key))

    # -----------------------------
    if not_equal:
        s_warnings += f'not equal ({len(not_equal)}): \n    ' + '\n    '.join(not_equal) + '\n'

    if not_in_1:
        s_warnings += f'not in inp1 ({len(not_in_1)}): ' + ' | '.join(not_in_1) + '\n'

    if not_in_2:
        s_warnings += f'not in inp2 ({len(not_in_2)}): ' + ' | '.join(not_in_2) + '\n'

    # -----------------------------
    res = ''
    if s_warnings:
        res += s_warnings
    else:
        res += 'good!\n'

    res += f'{n_equal}/{len(i1 | i2)} objects are equal\n'

    return res


def compare_inp_files(fn1, fn2, precision=2, skip_section=None):
    """
    compare two inp files and get the differences as string output

    Args:
        fn1 (str): filename for the first inp file
        fn2 (str): filename for the second inp file
        precision (int): number of relevant decimal places
        skip_section (list): skip sections if you don't care for specific changes

    Returns:
        str: differences of the files
    """
    s = f'Comparing \n' \
        f'   "{fn1}" (=inp1)\n' \
        f'   to\n' \
        f'   "{fn2}" (=inp2)\n\n'
    inp1 = SwmmInput.read_file(fn1)
    inp2 = SwmmInput.read_file(fn2)

    sections = set(inp1.keys()) | set(inp2.keys())

    for section in sorted(sections, key=_sort_by):
        if skip_section is not None and section in skip_section:
            continue
        if section in [TITLE]:
            continue
        s += '\n' + '#' * 100 + f'\n[{section}]\n'

        if (section in inp1) and (section in inp2):
            s += compare_sections(inp1[section], inp2[section], precision=precision)
        elif (section not in inp1) and (section in inp2):
            s += f'only in inp2\n'
        elif (section in inp1) and (section not in inp2):
            s += f'only in inp1\n'
        else:
            s += f'not in both inps\n'  # should not happen

    inp_version_control(inp1, inp2)

    return s


def _to_coordinates_list(geo_series):
    return list(zip(geo_series.x, geo_series.y))


def _next_nodes():
    # TODO ...
    coords_from = _to_coordinates_list(geometry_from.geometry)
    coords_to = _to_coordinates_list(geometry_to.geometry)
    # distance_matrix = distance.cdist(coords_from, coords_to, 'euclidean')
    # return np.argmin(distance_matrix, axis=1)

    from sklearn.neighbors import NearestNeighbors
    n_neighbors = 1
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree').fit(coords_to)
    distances, indices = nbrs.kneighbors(coords_from)


def inp_version_control(inp_v1, inp_v2):
    # zuerst neue knoten erstellen
    #   wurde nur konvertiert oder umbenannt?
    #

    # bestehende links zu neuen knoten umleiten
    # neue links einfügen
    # alte knoten löschen

    # ----
    g1 = inp_to_graph(inp_v1)
    g2 = inp_to_graph(inp_v2)
    # ----
    # NODES
    nodes1 = nodes_dict(inp_v1)
    nodes2 = nodes_dict(inp_v2)

    # Knoten wurde konvertiert
    for node in set(nodes2.keys()) & set(nodes1.keys()):
        if type(nodes1[node]) != type(nodes2[node]):
            print(f'Für Knoten "{node}" hat sich der Objekt-Typ geändert: "{type(nodes1[node])}" zu "{type(nodes2[node])}"')

    nodes_new = set(nodes2.keys()) - set(nodes1.keys())
    for node_new in nodes_new:
        # wurde umbenannt
        # gleiche koordinate
        # oder gleichen inflow und outflow
        previous_links(inp_v2, node_new, g=g2)
        next_links(inp_v2, node_new, g=g2)

        pass
    nodes_removed = set(nodes1.keys()) - set(nodes2.keys())

    # ----
    # LINKS
    links1 = links_dict(inp_v1)
    links2 = links_dict(inp_v2)
    links_removed = set(links1.keys()) - set(links2.keys())
    links_new = set(links2.keys()) - set(links1.keys())

    # Link wurde konvertiert
    for link in set(links1.keys()) & set(links2.keys()):
        if type(links1[link]) != type(links2[link]):
            print(f'Für Knoten "{link}" hat sich der Objekt-Typ geändert: "{type(links1[link])}" zu "{type(links2[link])}"')



import inspect
import time

from numpy import NaN
from pandas import MultiIndex
from tqdm.auto import tqdm

from .filter import filter_nodes, filter_links
from .geo import complete_vertices, simplify_vertices, reduce_vertices
from .tags import get_node_tags, get_link_tags, get_subcatchment_tags
from ..inp import SwmmInput
from ..section_labels import *
from ..section_lists import LINK_SECTIONS, NODE_SECTIONS
from ..section_types import SECTION_TYPES
from ..sections import Tag
from ..sections.map_geodata import (add_geo_support, InpSectionGeo, convert_section_to_geosection,
                                    geo_section_converter, CoordinateGeo, VerticesGeo, PolygonGeo, )

"""
{'AeronavFAA': 'r',
 'ARCGEN': 'r',
 'BNA': 'raw',
 'DXF': 'raw',
 'CSV': 'raw',
 'OpenFileGDB': 'r',
 'ESRIJSON': 'r',
 'ESRI Shapefile': 'raw',
 'GeoJSON': 'rw',
 'GPKG': 'rw',
 'GML': 'raw',
 'GPX': 'raw',
 'GPSTrackMaker': 'raw',
 'Idrisi': 'r',
 'MapInfo File': 'raw',
 'DGN': 'raw',
 'S57': 'r',
 'SEGY': 'r',
 'SUA': 'r',
 'TopoJSON': 'r'}
 
# end = '.json'
# end = '.shp'
# end = '.gpkg'
"""


def convert_inp_to_geo_package(inp_fn, gpkg_fn=None, driver='GPKG', label_sep='.', crs="EPSG:32633"):
    """
    convert inp file data from an .inp-file to a GIS database

    Args:
        inp_fn (str): filename of the SWMM inp file
        gpkg_fn (str): filename of the new geopackage file
        driver (str): The OGR format driver used to write the vector file. (see: fiona.supported_drivers)
        label_sep (str): separator for attribute label between section header and object attribute.
            I.e. "JUNCTIONS.Elevation" with label_sep='.'
        crs (str): Coordinate Reference System of the geometry objects.
                    Can be anything accepted by pyproj.CRS.from_user_input(),
                    such as an authority string (eg “EPSG:4326”) or a WKT string.
    """
    if gpkg_fn is None:
        gpkg_fn = inp_fn.replace('.inp', '.gpkg')

    inp = SwmmInput.read_file(inp_fn, custom_converter=geo_section_converter)

    write_geo_package(inp, gpkg_fn, driver=driver, label_sep=label_sep, crs=crs)


def write_geo_package(inp, gpkg_fn, driver='GPKG', label_sep='.', crs="EPSG:32633"):
    """
    write the inp file data to a GIS database

    Args:
        inp (SwmmInput): inp file data
        gpkg_fn (str): filename of the new geopackage file
        driver (str): The OGR format driver used to write the vector file. (see: fiona.supported_drivers)
        label_sep (str): separator for attribute label between section header and object attribute.
            I.e. "JUNCTIONS.Elevation" with label_sep='.'
        crs (str): Coordinate Reference System of the geometry objects.
                    Can be anything accepted by pyproj.CRS.from_user_input(),
                    such as an authority string (eg “EPSG:4326”) or a WKT string.
    """
    from geopandas import GeoDataFrame

    todo_sections = NODE_SECTIONS + LINK_SECTIONS + [SUBCATCHMENTS]
    print(*todo_sections, sep=' | ')

    add_geo_support(inp, crs=crs)

    # ---------------------------------
    t0 = time.perf_counter()
    nodes_tags = get_node_tags(inp)
    for sec in NODE_SECTIONS:
        if sec in inp:
            df = inp[sec].frame.rename(columns=lambda c: f'{sec}{label_sep}{c}')

            if sec == STORAGE:
                df[f'{STORAGE}{label_sep}Curve'] = df[f'{STORAGE}{label_sep}Curve'].astype(str)

            for sub_sec in [DWF, INFLOWS]:
                if sub_sec in inp:
                    x = inp[sub_sec].frame.unstack(1)
                    x.columns = [f'{label_sep}'.join([sub_sec, c[1], c[0]]) for c in x.columns]
                    df = df.join(x)

            df = df.join(inp[COORDINATES].geo_series).join(nodes_tags)

            GeoDataFrame(df).to_file(gpkg_fn, driver=driver, layer=sec)
            print(f'{f"{time.perf_counter() - t0:0.1f}s":^{len(sec)}s}', end=' | ')
        else:
            print(f'{f"-":^{len(sec)}s}', end=' | ')
        t0 = time.perf_counter()

    # ---------------------------------
    links_tags = get_link_tags(inp)
    complete_vertices(inp)
    simplify_vertices(inp)
    for sec in LINK_SECTIONS:
        if sec in inp:
            df = inp[sec].frame.rename(columns=lambda c: f'{sec}{label_sep}{c}').join(
                inp[XSECTIONS].frame.rename(columns=lambda c: f'{XSECTIONS}{label_sep}{c}'))

            if sec == OUTLETS:
                df[f'{OUTLETS}{label_sep}Curve'] = df[f'{OUTLETS}{label_sep}Curve'].astype(str)

            if LOSSES in inp:
                df = df.join(inp[LOSSES].frame.rename(columns=lambda c: f'{LOSSES}{label_sep}{c}'))

            df = df.join(inp[VERTICES].geo_series).join(links_tags)

            GeoDataFrame(df).to_file(gpkg_fn, driver=driver, layer=sec)

            print(f'{f"{time.perf_counter() - t0:0.1f}s":^{len(sec)}s}', end=' | ')
        else:
            print(f'{f"-":^{len(sec)}s}', end=' | ')
        t0 = time.perf_counter()

    # ---------------------------------
    if SUBCATCHMENTS in inp:
        df = inp[SUBCATCHMENTS].frame.rename(columns=lambda c: f'{SUBCATCHMENTS}{label_sep}{c}').join(
            inp[SUBAREAS].frame.rename(columns=lambda c: f'{SUBAREAS}{label_sep}{c}')).join(
            inp[INFILTRATION].frame.rename(columns=lambda c: f'{INFILTRATION}{label_sep}{c}')).join(
            inp[POLYGONS].geo_series).join(get_subcatchment_tags(inp))

        GeoDataFrame(df).to_file(gpkg_fn, driver=driver, layer=SUBCATCHMENTS)
        gs_connector = get_subcatchment_connectors(inp)
        GeoDataFrame(gs_connector).to_file(gpkg_fn, driver=driver, layer=SUBCATCHMENTS + '_connector')

        print(f'{f"{time.perf_counter() - t0:0.1f}s":^{len(SUBCATCHMENTS)}s}')
    else:
        print(f'{f"-":^{len(SUBCATCHMENTS)}s}', end=' | ')


def get_subcatchment_connectors(inp):
    """
    create connector line objects between subcatchment outlets and centroids

    Args:
        inp (SwmmInput): inp file data

    Returns:
        geopandas.GeoSeries: line objects between subcatchment outlets and centroids
    """
    # centroids = inp[s.POLYGONS].geo_series.centroid
    # outlets = inp[s.SUBCATCHMENTS].frame.Outlet
    # junctions = inp[s.COORDINATES].geo_series.reindex(outlets.values)
    # junctions.index = outlets.index
    from geopandas import GeoSeries
    from shapely.geometry import LineString

    res = dict()
    for p in tqdm(inp[POLYGONS], total=len(inp[POLYGONS].keys()), desc='get_subcatchment_connectors'):
        c = inp[POLYGONS][p].geo.centroid
        o = inp[SUBCATCHMENTS][p].Outlet
        if o not in inp[COORDINATES]:
            print(inp[SUBCATCHMENTS][p])
            continue
        res[p] = LineString([inp[COORDINATES][o].point, (c.x, c.y)])

    gs = GeoSeries(res, crs=inp[POLYGONS]._crs)
    gs.index.name = 'Subcatchment'
    gs.name = 'geometry'
    return gs


def problems_to_gis(inp, gpkg_fn, nodes=None, links=None, **kwargs):
    """
    filter inp data by nodes and links and write objects to a gis database

    Args:
        inp:
        gpkg_fn:
        nodes:
        links:
        **kwargs:
    """
    if nodes is not None:
        inp = filter_nodes(inp, nodes)

    if links is not None:
        inp = filter_links(inp, links)

    write_geo_package(inp, gpkg_fn, **kwargs)


def links_geo_data_frame(inp, label_sep='.'):
    """
    convert link data in inp file to geo-data-frame

    Args:
        inp (SwmmInput): inp file data
        label_sep (str): separator for attribute label between section header and object attribute.
            I.e. "JUNCTIONS.Elevation" with label_sep='.'

    Returns:
        geopandas.GeoDataFrame: links as geo-data-frame
    """
    from geopandas import GeoDataFrame

    if (VERTICES in inp) and not isinstance(inp[VERTICES], InpSectionGeo):
        inp[VERTICES] = convert_section_to_geosection(inp[VERTICES])
    links_tags = get_link_tags(inp)
    complete_vertices(inp)
    res = None
    for sec in LINK_SECTIONS:
        if sec in inp:
            df = inp[sec].frame.rename(columns=lambda c: f'{sec}{label_sep}{c}').join(
                inp[XSECTIONS].frame.rename(columns=lambda c: f'{XSECTIONS}{label_sep}{c}'))

            if sec == OUTLETS:
                df[f'{OUTLETS}{label_sep}Curve'] = df[f'{OUTLETS}{label_sep}Curve'].astype(str)

            if LOSSES in inp:
                df = df.join(inp[LOSSES].frame.rename(columns=lambda c: f'{LOSSES}{label_sep}{c}'))

            df = df.join(inp[VERTICES].geo_series).join(links_tags)
            if res is None:
                res = df
            else:
                res = res.append(df)
    return GeoDataFrame(res)


def nodes_geo_data_frame(inp, label_sep='.'):
    """
    convert nodes data in inp file to geo-data-frame

    Args:
        inp (SwmmInput): inp file data
        label_sep (str): separator for attribute label between section header and object attribute.
            I.e. "JUNCTIONS.Elevation" with label_sep='.'

    Returns:
        geopandas.GeoDataFrame: nodes as geo-data-frame
    """
    from geopandas import GeoDataFrame

    if (COORDINATES in inp) and not isinstance(inp[COORDINATES], InpSectionGeo):
        inp[COORDINATES] = convert_section_to_geosection(inp[COORDINATES])
    nodes_tags = get_node_tags(inp)
    res = None
    for sec in NODE_SECTIONS:
        if sec in inp:
            df = inp[sec].frame.rename(columns=lambda c: f'{sec}{label_sep}{c}')

            if sec == STORAGE:
                df[f'{STORAGE}{label_sep}Curve'] = df[f'{STORAGE}{label_sep}Curve'].astype(str)

            for sub_sec in [DWF, INFLOWS]:
                if sub_sec in inp:
                    x = inp[sub_sec].frame.unstack(1)
                    x.columns = [f'{label_sep}'.join([sub_sec, c[1], c[0]]) for c in x.columns]
                    df = df.join(x)

            df = df.join(inp[COORDINATES].geo_series).join(nodes_tags)

            if res is None:
                res = df
            else:
                res = res.append(df)

    return GeoDataFrame(res)


def gpkg_to_swmm(fn, label_sep='.'):
    """
    import inp data from GIS gpkg file created from the swmm_api.input_file.macro_snippets.gis_export.convert_inp_to_geo_package function

    Args:
        fn (str): filename to GPKG datebase file
        label_sep (str):  separator for attribute label between section header and object attribute.
            I.e. "JUNCTIONS.Elevation" with label_sep='.'

    Returns:
        SwmmInput: inp data
    """
    import fiona
    from geopandas import read_file

    inp = SwmmInput()

    SECTION_TYPES.update({COORDINATES: CoordinateGeo,
                          VERTICES   : VerticesGeo,
                          POLYGONS   : PolygonGeo})

    def _check_sec(sec):
        if sec not in inp:
            inp[sec] = SECTION_TYPES[sec].create_section()

    for sec in NODE_SECTIONS:
        if sec not in fiona.listlayers(fn):
            continue
        gdf = read_file(fn, layer=sec).set_index('Name')

        cols = gdf.columns[gdf.columns.str.startswith(sec)]
        inp[sec] = SECTION_TYPES[sec].create_section(gdf[cols].reset_index().fillna(NaN).values)

        for sub_sec in [DWF, INFLOWS]:
            _check_sec(sub_sec)
            cols = gdf.columns[gdf.columns.str.startswith(sub_sec)]
            gdf_sub = gdf[cols].copy()
            gdf_sub.columns = MultiIndex.from_tuples([col.split(label_sep) for col in gdf_sub.columns])
            cols_order = gdf_sub.columns.droplevel(1)
            gdf_sub = gdf_sub.stack(1)[cols_order]
            inp[sub_sec].update(SECTION_TYPES[sub_sec].create_section(gdf_sub.reset_index().values))

        _check_sec(COORDINATES)
        inp[COORDINATES].update(SECTION_TYPES[COORDINATES].create_section_from_geoseries(gdf.geometry))

        tags = gdf[['tag']].copy()
        tags['type'] = Tag.TYPES.Node
        _check_sec(TAGS)
        inp[TAGS].update(SECTION_TYPES[TAGS].create_section(tags[['type', 'tag']].reset_index().values))

    for i in inp[STORAGE]:
        c = inp[STORAGE][i].Curve
        if isinstance(c, list):
            inp[STORAGE][i].Curve = [float(j) for j in c[0][1:-1].split(',')]

    # ---------------------------------
    for sec in LINK_SECTIONS:
        if sec not in fiona.listlayers(fn):
            continue
        gdf = read_file(fn, layer=sec).set_index('Name')

        cols = gdf.columns[gdf.columns.str.startswith(sec)]
        inp[sec] = SECTION_TYPES[sec].create_section(gdf[cols].reset_index().fillna(NaN).values)

        for sub_sec in [XSECTIONS, LOSSES]:
            _check_sec(sub_sec)
            cols = gdf.columns[gdf.columns.str.startswith(sub_sec)].to_list()
            if cols:
                if sub_sec == XSECTIONS:
                    cols = [f'{sub_sec}{label_sep}{i}' for i in inspect.getargspec(SECTION_TYPES[sub_sec]).args[2:]]
                gdf_sub = gdf[cols].copy().dropna(how='all')
                if sub_sec == LOSSES:
                    gdf_sub[f'{LOSSES}{label_sep}FlapGate'] = gdf_sub[f'{LOSSES}{label_sep}FlapGate'] == 1
                inp[sub_sec].update(SECTION_TYPES[sub_sec].create_section(gdf_sub.reset_index().values))

        _check_sec(VERTICES)
        inp[VERTICES].update(SECTION_TYPES[VERTICES].create_section_from_geoseries(gdf.geometry))

        tags = gdf[['tag']].copy()
        tags['type'] = Tag.TYPES.Link
        _check_sec(TAGS)
        inp[TAGS].update(SECTION_TYPES[TAGS].create_section(tags[['type', 'tag']].reset_index().values))

    if OUTLETS in inp:
        for i in inp[OUTLETS]:
            c = inp[OUTLETS][i].Curve
            if isinstance(c, list):
                inp[OUTLETS][i].Curve = [float(j) for j in c[0][1:-1].split(',')]

    simplify_vertices(inp)
    reduce_vertices(inp)

    # ---------------------------------
    if SUBCATCHMENTS in fiona.listlayers(fn):
        gdf = read_file(fn, layer=SUBCATCHMENTS).set_index('Name')

        for sec in [SUBCATCHMENTS, SUBAREAS, INFILTRATION]:
            cols = gdf.columns[gdf.columns.str.startswith(sec)]
            inp[sec] = SECTION_TYPES[sec].create_section(gdf[cols].reset_index().fillna(NaN).values)

        inp[POLYGONS] = SECTION_TYPES[POLYGONS].create_section_from_geoseries(gdf.geometry)

    return inp


def update_length(inp):
    """
    Set the length of all conduits based on the length of the vertices.

    Args:
        inp (SwmmInput): inp data

    .. Important::
        Works inplace.
    """
    inp[VERTICES] = convert_section_to_geosection(inp[VERTICES], crs="EPSG:32633")
    for c in inp.CONDUITS.values():
        c.Length = inp.VERTICES[c.Name].geo.length



"""Predefined input file section header to prevent typos"""

TITLE = 'TITLE'  # project title
OPTIONS = 'OPTIONS'  # analysis options
REPORT = 'REPORT'  # output reporting instructions
FILES = 'FILES'  # interface file options
RAINGAGES = 'RAINGAGES'  # rain gage information
EVAPORATION = 'EVAPORATION'  # evaporation data
TEMPERATURE = 'TEMPERATURE'  # air temperature and snow melt data
ADJUSTMENTS = 'ADJUSTMENTS'  # monthly adjustments applied to climate variables

SUBCATCHMENTS = 'SUBCATCHMENTS'  # basic subcatchment information
SUBAREAS = 'SUBAREAS'  # subcatchment impervious/pervious sub-area data
INFILTRATION = 'INFILTRATION'  # subcatchment infiltration parameters
LID_CONTROLS = 'LID_CONTROLS'  # low impact development control information
LID_USAGE = 'LID_USAGE'  # assignment of LID controls to subcatchments
AQUIFERS = 'AQUIFERS'  # groundwater aquifer parameters
GROUNDWATER = 'GROUNDWATER'  # subcatchment groundwater parameters
GWF = 'GWF'  # groundwater flow expressions
SNOWPACKS = 'SNOWPACKS'  # subcatchment snow pack parameters

JUNCTIONS = 'JUNCTIONS'  # junction node information
OUTFALLS = 'OUTFALLS'  # outfall node information
DIVIDERS = 'DIVIDERS'  # flow divider node information
STORAGE = 'STORAGE'  # storage node information

CONDUITS = 'CONDUITS'  # conduit link information
PUMPS = 'PUMPS'  # pump link information
ORIFICES = 'ORIFICES'  # orifice link information
WEIRS = 'WEIRS'  # weir link information
OUTLETS = 'OUTLETS'  # outlet link information

XSECTIONS = 'XSECTIONS'  # conduit, orifice, and weir cross-section geometry
TRANSECTS = 'TRANSECTS'  # transect geometry for conduits with irregular cross-sections
LOSSES = 'LOSSES'  # conduit entrance/exit losses and flap valves

CONTROLS = 'CONTROLS'  # rules that control pump and regulator operation

POLLUTANTS = 'POLLUTANTS'  # pollutant information
LANDUSES = 'LANDUSES'  # land use categories
COVERAGES = 'COVERAGES'  # assignment of land uses to subcatchments
LOADINGS = 'LOADINGS'  # initial pollutant loads on subcatchments
BUILDUP = 'BUILDUP'  # buildup functions for pollutants and land uses
WASHOFF = 'WASHOFF'  # washoff functions for pollutants and land uses

TREATMENT = 'TREATMENT'  # pollutant removal functions at conveyance system nodes
INFLOWS = 'INFLOWS'  # external hydrograph/pollutograph inflow at nodes
DWF = 'DWF'  # baseline dry weather sanitary inflow at nodes
RDII = 'RDII'  # rainfall-dependent I/I information at nodes

HYDROGRAPHS = 'HYDROGRAPHS'  # unit hydrograph data used to construct RDII inflows
CURVES = 'CURVES'  # x-y tabular data referenced in other sections
TIMESERIES = 'TIMESERIES'  # time series data referenced in other sections
PATTERNS = 'PATTERNS'  # periodic multipliers referenced in other sections

# ___________________________
# NOT IN SWMM DOCU

TAGS = 'TAGS'

COORDINATES = 'COORDINATES'  # node  # X,Y coordinates for nodes
VERTICES = 'VERTICES'  # link  # X,Y coordinates for each interior vertex of polyline links
POLYGONS = 'POLYGONS'  # catchment  # X,Y coordinates for each vertex of subcatchment polygons

MAP = 'MAP'  # view  # X,Y coordinates of the map’s bounding rectangle
LABELS = 'LABELS'  # free text  # X,Y coordinates and text of labels
SYMBOLS = 'SYMBOLS'  # rain gauge  # X,Y coordinates for rain gages
BACKDROP = 'BACKDROP'  # background pic # X,Y coordinates of the bounding rectangle and file name of the backdrop image.

PROFILES = 'PROFILES'  # Profile plot nodes



from .check import check_for_nodes, check_for_duplicates
from .collection import nodes_dict, links_dict, subcatchments_per_node_dict, nodes_subcatchments_dict
from .compare import compare_sections, compare_inp_files
from .convert import junction_to_storage, junction_to_outfall, conduit_to_orifice
from .curve import curve_figure
from .edit import (combine_conduits, combine_conduits_keep_slope, combine_vertices, delete_link, delete_node,
                   delete_subcatchment, dissolve_conduit, flip_link_direction, move_flows, rename_link, rename_node,
                   rename_subcatchment, rename_timeseries, split_conduit, remove_quality_model)
from .filter import (filter_tags, filter_nodes, filter_links_within_nodes, filter_links, filter_subcatchments,
                     create_sub_inp, )
try:
    import warnings

    warnings.filterwarnings('ignore', message='.*is incompatible with the GEOS version PyGEOS*')

    from .geo import (transform_coordinates, complete_vertices, reduce_vertices, complete_link_vertices,
                      simplify_link_vertices, simplify_vertices, )
    from .gis import (convert_inp_to_geo_package, write_geo_package, get_subcatchment_connectors,
                      links_geo_data_frame, nodes_geo_data_frame, gpkg_to_swmm, update_length, problems_to_gis)
except ImportError as e:
    print('Needed Packages: pyproj, fiona, geopandas, shapely')
    print(e)
    pass

from .graph import (inp_to_graph, get_path, get_path_subgraph, next_links, next_links_labels, next_nodes,
                    previous_links, previous_links_labels, previous_nodes, links_connected, number_in_out,
                    downstream_nodes, upstream_nodes, get_network_forks, split_network, conduit_iter_over_inp)
from .macros import (find_node, find_link, calc_slope, conduit_slopes, conduits_are_equal, update_no_duplicates,
                     increase_max_node_depth, set_times, combined_subcatchment_frame, delete_sections)
from .plotting import plot_map, plot_longitudinal
from .reduce_unneeded import (reduce_curves, reduce_controls, simplify_curves, reduce_raingages,
                              remove_empty_sections, )
from .split_inp_file import split_inp_to_files, read_split_inp_file
from .tags import get_node_tags, get_link_tags, get_subcatchment_tags, filter_tags



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import subprocess
import os
from sys import platform as _platform


class SWMMRunError(UserWarning):
    pass


def get_swmm_command_line(swmm_path, inp, rpt, out):
    cmd = (swmm_path, inp, rpt, out)
    return cmd, inp, rpt, out


def infer_swmm_path():
    # UNIX
    swmm_path = 'swmm5'

    # WINDOWS
    if _platform.startswith("win"):
        swmm_path = None
        # script_path = '???/swmm5.exe'
        for program_files in ['Program Files (x86)', 'Program Files']:
            for version in ['5.1.015', '5.1.014', '5.1.013']:
                script_path = os.path.join('C:\\', program_files, 'EPA SWMM {}'.format(version), 'swmm5.exe')
                if os.path.isfile(script_path):
                    swmm_path = script_path
                    break
            if swmm_path is not None:
                break

    return swmm_path


def get_result_filenames(inp_fn):
    """
    get filenames for the Report and Output files

    Args:
        inp_fn (str): filename of the Input-Files

    Returns:
        tuple(str, str): filenames for the Report- and Output-file
    """
    return inp_fn.replace('.inp', '.rpt'), inp_fn.replace('.inp', '.out')


def get_swmm_command_line_auto(inp, rpt_dir=None, out_dir=None, create_out=True, swmm_path=None):
    base_filename = os.path.basename(inp).replace('.inp', '')
    inp_dir = os.path.dirname(inp)

    # -----------------------
    if rpt_dir is None:
        rpt_dir = inp_dir

    rpt = os.path.join(rpt_dir, base_filename + '.rpt')

    # -----------------------
    if out_dir is None:
        out_dir = inp_dir

    if create_out:
        out = os.path.join(out_dir, base_filename + '.out')
    else:
        out = ''

    # -----------------------
    if swmm_path is None:
        swmm_path = infer_swmm_path()

    return get_swmm_command_line(swmm_path, inp, rpt, out)


def run_swmm_stdout(command_line, sep='_' * 100):
    print(sep)
    print(command_line)
    subprocess.run(command_line)
    print(sep)


def run_swmm_custom(command_line):
    # shell_output = subprocess.run(command_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    shell_output = subprocess.run(command_line, capture_output=True)
    return shell_output


def check_swmm_errors(rpt, shell_output):
    msgs = dict()

    if isinstance(shell_output, str):
        msgs['CALL'] = shell_output

    else:
        stdout = shell_output.stdout.decode()
        # if 'error' in stdout:
        msgs.update({
            'CALL': shell_output.args,
            'RETURN': shell_output.returncode,
            'ERROR': shell_output.stderr.decode(),
            'OUT': stdout,
        })

    if os.path.isfile(rpt):
        with open(rpt, 'r') as f:
            rpt_content = f.read()
        if 'ERROR' in rpt_content:
            msgs['REPORT'] = rpt_content
    else:
        msgs['REPORT'] = 'NO Report file created!!!'

    if 'REPORT' in msgs:
        sep = '\n' + '_' * 100 + '\n'
        error_msg = sep + sep.join('{}:\n  {}'.format(k, v) for k, v in msgs.items())
        raise SWMMRunError(error_msg)


def swmm5_run(inp, rpt_dir=None, out_dir=None, init_print=False, create_out=True, swmm_path=None):
    """
    run a simulation with an EPA-SWMM input-file
    default working directory is input-file directory

    Args:
        inp (str): path to input file
        rpt_dir (str): directory in which the report-file is written.
        out_dir (str): directory in which the output-file is written.
        init_print (bool): if the default commandline output should be printed
        create_out (bool): if the out-file should be created

    Returns:
        tuple[str, str, str]: INP-, RPT- and OUT-filename
    """
    command_line, inp, rpt, out = get_swmm_command_line_auto(inp, rpt_dir=rpt_dir, out_dir=out_dir,
                                                             create_out=create_out, swmm_path=swmm_path)
    # -------------------------
    if init_print:
        run_swmm_stdout(command_line)
        stdout = ' '.join(command_line)
    else:
        stdout = run_swmm_custom(command_line)

    # -------------------------
    check_swmm_errors(rpt, stdout)

    return rpt, out


def swmm5_run_parallel(inp_fns, processes=4):
    """
    run multiple swmm models in parallel

    Args:
        inp_fns (list): list of SWMM modell filenames (.inp-files)
        processes (int): number of parallel processes
    """
    _run_parallel(inp_fns, swmm5_run, processes=processes)


def _run_parallel(variable, func=swmm5_run, processes=4):
    from tqdm.auto import tqdm
    from functools import partial

    if processes == 1:
        for fn_inp in tqdm(variable):
            func(fn_inp)

    else:
        from multiprocessing.dummy import Pool

        pool = Pool(processes)
        for _ in tqdm(pool.imap(partial(func), variable), total=len(variable)):
            pass


def get_swmm_version():
    swmm_path = infer_swmm_path()
    shell_output = subprocess.run([swmm_path, '--version'], capture_output=True)
    return shell_output.stdout.decode().strip()



from .section_labels import *

GEO_SECTIONS = [COORDINATES, VERTICES, POLYGONS]
GUI_SECTIONS = [MAP, SYMBOLS, LABELS, BACKDROP, PROFILES]
LINK_SECTIONS = [CONDUITS, ORIFICES, OUTLETS, PUMPS, WEIRS]
NODE_SECTIONS = [DIVIDERS, JUNCTIONS, OUTFALLS, STORAGE]
SUBCATCHMENT_SECTIONS = [SUBCATCHMENTS, SUBAREAS, INFILTRATION, POLYGONS]
POLLUTANT_SECTIONS = [TREATMENT, WASHOFF, BUILDUP, POLLUTANTS]



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from io import StringIO
import pandas as pd
import re
from numpy import NaN


def _get_title_of_part(part, alt):
    """
    read title in part text of the report file

    Args:
        part (str): string of a part in the report file
        alt: alternative title

    Returns:
        str: title of the part
    """
    if 'EPA STORM WATER MANAGEMENT MODEL - VERSION' in part:
        return 'Version+Title'

    elif 'NOTE:' in part:
        return 'Note'

    else:
        lines = part.split('\n')
        for no, line in enumerate(lines):
            if no == 0 or no == len(lines) - 1:
                continue
            if '***' in lines[no + 1] and '***' in lines[no - 1]:
                start_char = lines[no - 1].index('*')
                len_title = lines[no - 1].count('*')
                return line[start_char:start_char + len_title].strip()
        return str(alt)


def _remove_lines(part, title=True, empty=False, sep=False):
    """
    remove unneeded lines of part string

    Args:
        part (str): part in report file
        title (bool): if title should be removed
        empty (bool): if empty lines should be removed
        sep (bool):  if separator should be removed

    Returns:
        str: converted part
    """
    lines = part.split('\n')
    remove_lines = list()
    for no, line in enumerate(lines):

        if title:
            if no == 0 or no == len(lines) - 1:
                continue
            if '***' in lines[no + 1] and '***' in lines[no - 1]:
                remove_lines.append(no - 1)
                remove_lines.append(no)
                remove_lines.append(no + 1)

        if empty:
            if len(line.strip()) == 0:
                remove_lines.append(no)

        if sep:
            if len(line.replace('-', '').strip()) == 0:
                remove_lines.append(no)

    new_lines = lines
    for r in sorted(remove_lines, reverse=True):
        new_lines.pop(r)

    return '\n'.join(new_lines)


def _part_to_frame(part):
    """
    convert the table of a part of the report file to a dataframe

    Args:
        part (str): part of the report file

    Returns:
        pandas.DataFrame: some kind of summary table
    """
    subs = re.split(r"\s*-+\n", part)
    if len(subs) == 4:
        notes, header, data, sytem = subs
    elif len(subs) == 1:
        # no data
        return

    elif len(subs) == 2:
        # input summary tables
        header, data = subs
    else:
        notes, header, data = subs
    header = ['_'.join([i for i in c if i is not NaN]) for c in pd.read_fwf(StringIO(header), header=None).values.T]

    # Pumping Summary
    if '% Time Off_Pump Curve_Low   High' in header:
        header.remove('% Time Off_Pump Curve_Low   High')
        header.append('% Time Off_Pump Curve_Low')
        header.append('% Time Off_Pump Curve_High')

    df = pd.DataFrame(line.split() for line in data.split('\n'))

    last_col_values = df.iloc[:, -1].unique()
    if 'ltr' in last_col_values or 'gal' in last_col_values:
        del df[df.columns[-1]]

    for c, h in enumerate(header):
        if (('days hr:min' in h)
                or ('Recording_Frequency' in h)
                or ('Interval' in h)
                or ('CoPollutant' in h)):
            if c+1 not in df.columns:
                continue
            df[c] = df[c] + ' ' + df.pop(c+1)

    if len(df.columns) < len(header):
        df.columns = header[:len(df.columns)]
        for h in header[len(df.columns):]:
            df[h] = NaN
    else:
        df.columns = header

    df = df.set_index(header[0])

    # New error in Version 5.1.015 ????
    df = df.replace('-nan(ind)', NaN)

    for col in df:
        if 'days hr:min' in col:
            df[col] = pd.to_timedelta(df[col].str.replace('  ', ' days ') + ':00')
        else:
            df[col] = pd.to_numeric(df[col], errors='ignore')

    return df.copy()


def _routing_part_to_dict(raw):
    elements = dict()
    if (raw.startswith('  All ') and raw.endswith('.')) or (raw.strip() == 'None') or (raw.strip() == ''):
        return elements

    for line in raw.split('\n'):
        line = line.split()
        elements[line[1]] = line[-1][1:-1]
    return elements


def _continuity_part_to_dict(raw):
    if raw is None:
        return dict()

    df = pd.read_fwf(StringIO(raw), index_col=0, header=[0, 1, 2])

    df.columns = df.columns.droplevel(2)
    df.columns.name = None
    df.columns = ['_'.join(str(c) for c in col).strip() for col in df.columns.values]

    df.index.name = None
    df.index = df.index.str.replace('.', '', regex=False).str.strip()

    res = df.to_dict(orient='index')
    res['Continuity Error (%)'] = list(res['Continuity Error (%)'].values())[0]

    return res


class ReportUnitConversion:
    """
    Unit conversion for the simulation results in the report file
    
    Attributes:
        FLOW (str): for metric one of ['CMS', 'LPS', 'MLD'] | for imperial one of ['CFS', 'GPM', 'MGD']
        VOL1 (str): for metric: 'hectare-m' | for imperial:
        VOL2 (str): for metric: 'ltr' | for imperial: 'acre-feet'
        DEPTH1 (str): for metric: 'mm'| for imperial: 'gal'
        DEPTH2 (str): hydrological | for metric: 'Meters' | for imperial: 'inches'
        MASS (str): hydraulic | for metric: 'kg' | for imperial: 'Feet'
        LENGTH (str): for metric: 'm' | for imperial: 'lbs'
        VOL3 (str): for metric: 'm3' | for imperial: 'ft3'
        VELO (str): for metric: 'm/sec' | for imperial: 'ft/sec'

    Examples
    --------
    >>> ReportUnitConversion('CMS').is_metric()
    True
    >>> ReportUnitConversion('GPM').is_imperial()
    ['CFS', 'GPM', 'MGD']
    >>> ReportUnitConversion('CMS').DEPTH1
    'mm'
    """
    _METRIC_FLOWS = ['CMS', 'LPS', 'MLD']
    _IMPERIAL_FLOWS = ['CFS', 'GPM', 'MGD']

    def __init__(self, flow_unit):
        """
        Unit conversion

        Args:
            flow_unit (str): unit of simulation results in the report file
        """
        self.FLOW = flow_unit
        if self.is_metric():
            self.VOL1 = 'hectare-m'
            self.VOL2 = 'ltr'
            self.DEPTH1 = 'mm'  # hydrological
            self.DEPTH2 = 'Meters'  # hydraulic
            self.MASS = 'kg'
            self.LENGTH = 'm'
        else:
            self.VOL1 = 'acre-feet'
            self.VOL2 = 'gal'
            self.DEPTH1 = 'inches'  # hydrological
            self.DEPTH2 = 'Feet'  # hydraulic
            self.MASS = 'lbs'
            self.LENGTH = 'ft'

        self.VOL3 = self.LENGTH + '3'
        self.VELO = self.LENGTH + '/sec'

    def is_metric(self):
        """
        Indicator whether unit is metric.

        Returns:
            bool: If unit is metric.

        See Also:
            UNIT.is_imperial : If unit is imperial.
        """
        return self.FLOW in self._METRIC_FLOWS

    def is_imperial(self):
        """
        Indicator whether unit is imperial.

        Returns:
            bool: If unit is imperial.

        See Also:
            UNIT.is_metric : If unit is metric.
        """
        return self.FLOW in self._IMPERIAL_FLOWS


class ContinuityVariables:
    def __init__(self, flow_unit):
        """
        Unit conversion

        Args:
            flow_unit (str): unit of simulation results in the report file
        """
        u = ReportUnitConversion(flow_unit)
        self.VOL_HM3 = f'Volume_{u.VOL1}'
        self.VOL_1e6L = f'Volume_10^6 {u.VOL2}'
        self.DEPTH_MM = f'Depth_{u.DEPTH1}'



from bokeh.plotting import figure, output_file, show, save
import matplotlib.pyplot as plt
from matplotlib import patches
from shapely import geometry as shp

from ..macros import find_link
from ..sections import Outfall, Polygon, SubCatchment
from swmm_api.input_file.section_labels import *


def plot_map(inp):  # TODO
    fig = figure(
        tools="pan,box_zoom,reset,save",
        title="SWMM Model",
        # y_axis_type="log",
        # y_range=[0.001, 10 ** 11],
        # x_axis_label='sections',
        # y_axis_label='particles'
        x_axis_location=None, y_axis_location=None, sizing_mode='stretch_height'
    )
    fig.aspect_ratio = 1
    fig.grid.grid_line_color = None
    fig.background_fill_color = "#eeeeee"
    fig.axis.visible = False
    fig.y_range.flipped = True

    def _points(c):
        return c.x, c.y

    for section in [CONDUITS,
                    PUMPS,
                    ORIFICES,
                    WEIRS,
                    OUTLETS]:
        if section in inp:
            for link in inp[section].values():
                if link.Name in inp[VERTICES]:
                    points = [_points(inp[COORDINATES][link.FromNode])] \
                             + inp[VERTICES][link.Name].vertices \
                             + [_points(inp[COORDINATES][link.ToNode])]
                else:
                    points = [_points(inp[COORDINATES][link.FromNode]), _points(inp[COORDINATES][link.ToNode])]

                x, y = zip(*points)
                fig.line(x, y, line_color="black", line_width =1)
                fig.line(x, y, line_color="yellow", line_width =0.5)

    if POLYGONS in inp:
        for poly in inp[POLYGONS].values():  # type: Polygon
            x, y = zip(*poly.polygon)
            fig.patch(x, y)
            # ax.plot(x, y, 'r-')
            center = shp.Polygon(poly.polygon).centroid

            fig.scatter(x=center.x, y=center.y, marker='square', fill_color='black')

            subcatch = inp[SUBCATCHMENTS][poly.Subcatch]  # type: SubCatchment
            outlet = subcatch.Outlet
            outlet_point = inp[COORDINATES][outlet]
            fig.line([center.x, outlet_point.x], [center.y, outlet_point.y], line_color='red', line_dash="dashed")

    coords = inp[COORDINATES].frame
    node_style = {
        JUNCTIONS: {'marker': 'circle', 'color': 'blue'},
        STORAGE: {'marker': 'square', 'color': 'green'},
        OUTFALLS: {'marker': 'triangle', 'color': 'red'},

    }
    fig.scatter(x=coords.x, y=coords.y, marker=node_style[JUNCTIONS]['marker'], fill_color=node_style[JUNCTIONS]['color'], line_color='black')

    for section in [STORAGE, OUTFALLS]:
        if section in inp:
            is_in_sec = coords.index.isin(inp[section].keys())
            fig.scatter(coords[is_in_sec].x, coords[is_in_sec].y, marker=node_style[section]['marker'], size=5,
                        line_color="black", fill_color=node_style[section]['color'], alpha=0.5)

    return fig



import datetime
import re
from datetime import date, time, timedelta
from pandas import isna, to_datetime, Timedelta, Timestamp, to_timedelta
from pandas.tseries.frequencies import to_offset
from numpy import ndarray, NaN


def to_bool(x):
    if isinstance(x, bool):
        return x
    else:
        if x == 'YES':
            return True
        elif x == 'NO':
            return False
        else:
            raise NotImplemented('x not a bool: {}'.format(x))


def infer_type(x):
    """
    infer generic type of inp-file-string

    Args:
        x (str | list[str]):

    Returns:
        object: object depending on string
    """
    if isinstance(x, (list, ndarray)):
        return [infer_type(i) for i in x]
    elif not isinstance(x, str):
        return x
    elif x == 'YES':
        return True
    elif x == 'NO':
        return False
    elif x == 'NONE':
        return None
    elif x.replace('-', '').isdecimal():
        return int(x)
    elif ('.' in x) and (x.lower().replace('.', '').replace('-', '').replace('e', '').isdecimal()):
        return float(x)
    elif x.count('/') == 2:
        return to_datetime(x, format='%m/%d/%Y').date()
    # elif x.count('/') == 1:
    #     return to_datetime(x, format='%m/%d').date()
    elif (x.count(':') == 2) and (len(x) == 8):
        return to_datetime(x, format='%H:%M:%S').time()
    elif (x.count(':') == 1) and (len(x) == 5):
        return to_datetime(x, format='%H:%M').time()
    else:
        return x


def str_to_datetime(date=None, time=None, str_only=False):
    if date:
        if '-' in date:
            date = date.replace('-', '/')

        month = date.split('/')[0]
        if len(month) <= 2:
            month_format = '%m'
        elif len(month) == 3:
            month_format = '%b'
        else:
            raise NotImplementedError(date)

        if date.count('/') == 2:
            date_format2 = '/%d/%Y'
        else:
            raise NotImplementedError(date)
    else:
        date = ''
        month_format = ''
        date_format2 = ''

    if time:
        if date == '':
            parts = time.split(':')
            if len(parts) == 1:
                return float(parts[0])
            elif len(parts) == 2:
                return float(parts[0]) + float(parts[1])/60
            elif len(parts) == 3:
                return float(parts[0]) + float(parts[1])/60 + float(parts[2])/60/60
        else:
            time_format = '%H:%M:%S'
            if time.count(':') == 1:
                # time_format = '%H:%M'
                time += ':00'
            elif time.count(':') == 2:
                pass
                # time_format = '%H:%M:%S'
            elif time.count(':') == 0:
                hours = float(time)
                h = int(hours)
                minutes = (hours - h)*60
                m = int(minutes)
                s = int((minutes - m)*60)
                time = f'{h:02d}:{m:02d}:{s:02d}'

            else:
                raise NotImplementedError(time)

    else:
        time = ''
        time_format = ''

    if str_only:
        return date + ' ' + time
    else:
        return to_datetime(date + ' ' + time,
                           format=month_format + date_format2 + ' ' + time_format)


def datetime_to_str(dt):
    if isinstance(dt, float):
        hours = dt
        h = int(round(hours, 4))
        minutes = (hours - h) * 60
        m = int(round(minutes, 4))
        t = f'{h:02d}:{m:02d}'
        second = (minutes - m) * 60
        s = int(round(second, 4))
        if s:
            t += f':{s:02d}'
        return t
    elif isinstance(dt, str):
        return dt
    elif isinstance(dt, datetime.datetime):
        return dt.strftime(format='%m/%d/%Y %H:%M:%S')



def time2delta(t):
    return Timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)


def delta2time(d):
    return Timestamp(2000, 1, 1, d.components.hours, d.components.minutes, d.components.seconds).time()


def delta2offset(d):
    return to_offset(d)


def offset2delta(o):
    return to_timedelta(o)


def delta2str(d):
    """

    Args:
        d (pandas.Timedelta):

    Returns:
        str: HH:MM:SS
    """
    hours, remainder = divmod(d.total_seconds(), 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02.0f}:{:02.0f}:{:02.0f}'.format(hours, minutes, seconds)


def type2str(x):
    """
    convert any type to a string

    Args:
        x (any):

    Returns:
        str:
    """
    if isinstance(x, str):
        if ' ' in x:
            return f'"{x}"'
        return x
    elif isinstance(x, list):
        return ' '.join([type2str(i) for i in x])
    elif isinstance(x, bool):
        return 'YES' if x else 'NO'
    elif x is None:
        return 'NONE'
    elif isinstance(x, int):
        return str(x)
    elif isinstance(x, float):
        if isna(x):
            return ''
        if x == 0.0:
            return '0'
        return '{:0.7G}'.format(x)  # .rstrip('0').rstrip('.')
    elif isinstance(x, date):
        return x.strftime(format='%m/%d/%Y')
    elif isinstance(x, time):
        return x.strftime(format='%H:%M:%S')
    elif isinstance(x, (Timedelta, timedelta)):
        return delta2str(x)
    else:
        return str(x)

import numpy as np


def is_equal(x1, x2, precision=3):
    if isinstance(x1, float) and np.isnan(x1) and isinstance(x2, float) and np.isnan(x2):
        return True
    else:
        if isinstance(x1, float):
            x1 = round(x1, precision)
        if isinstance(x2, float):
            x2 = round(x2, precision)
        return x1 == x2


def convert_string(x) -> str:
    if isna(x):
        return x
    x = str(x)
    if ' ' in x:
        return f'"{x}"'
    else:
        s = x.strip('"')
        if s == '':
            return NaN
        else:
            return s


GIS_FLOAT_FORMAT = '0.03f'

_SECTION_PATTERN = re.compile(r'^[ \t]*([^;\n]+)[ \t]*;?[^\n]*$', flags=re.M)


def txt_to_lines(content):
    """
    converts text to multiple line arguments

    Comments will be ignored:
        ;; section comment
        ; object comment / either inline(at the end of the line) or before the line

    Args:
        content (str): section text
    Yields:
        list[str]: arguments per line in the input file section
    """
    for line in _SECTION_PATTERN.finditer(content):
        yield line.group().split()



from numpy import NaN, isnan

from ._identifiers import IDENTIFIERS
from .._type_converter import convert_string, GIS_FLOAT_FORMAT
from ..helpers import BaseSectionObject
from ..section_labels import DWF, INFLOWS, COORDINATES, RDII, TREATMENT


class DryWeatherFlow(BaseSectionObject):
    """
    Section: [**DWF**]

    Purpose:
        Specifies dry weather flow and its quality entering the drainage system at specific nodes.

    Format:
        ::

            Node Type Base (Pat1 Pat2 Pat3 Pat4)

    Formats-PCSWMM:
        ``Node Parameter AverageValue TimePatterns``

    Formats-SWMM-GUI:
        ``Node Constituent Baseline Patterns``

    Remarks:
        Pat1, Pat2, etc.
            names of up to four time patterns appearing in the [``PATTERNS``] section.

        The actual dry weather input will equal the product of the baseline value and any adjustment factors
        supplied by the specified patterns. (If not supplied, an adjustment factor defaults to 1.0.)
        The patterns can be any combination of monthly, daily, hourly and weekend hourly
        patterns, listed in any order. See the [PATTERNS] section for more details.

    Args:
        Node (str): name of node where dry weather flow enters.
        Constituent (str): keyword ``FLOW`` for flow or pollutant name for quality constituent. ``Type``
        Base (float): average baseline value for corresponding constituent (flow or concentration units).
        pattern1 (str, Optional): i.e.: monthly-pattern ``Pat1``
        pattern2 (str, Optional): i.e.: daily-pattern ``Pat2``
        pattern3 (str, Optional): i.e.: hourly-pattern
        pattern4 (str, Optional): i.e.: weekend-hourly-pattern
    """
    _identifier = (IDENTIFIERS.Node, IDENTIFIERS.Constituent)
    _section_label = DWF

    class TYPES:
        FLOW = 'FLOW'

    def __init__(self, Node, Constituent, Base, pattern1=NaN, pattern2=NaN, pattern3=NaN, pattern4=NaN, *patternx):
        self.Node: str = str(Node)
        """name of node where dry weather flow enters."""

        self.Constituent: str = Constituent
        """keyword ``FLOW`` for flow or pollutant name for quality constituent. ``Type``"""

        self.Base: float = float(Base)
        """average baseline value for corresponding constituent (flow or concentration units)."""

        self.pattern1: str = convert_string(pattern1)
        """i.e.: monthly-pattern ``Pat1``"""

        self.pattern2: str = convert_string(pattern2)
        """i.e.: daily-pattern ``Pat2``"""

        self.pattern3: str = convert_string(pattern3)
        """i.e.: hourly-pattern"""

        self.pattern4: str = convert_string(pattern4)
        """i.e. weekend-hourly-pattern"""

    def get_pattern_list(self):
        return [self[p] for p in ['pattern1', 'pattern2', 'pattern3', 'pattern4'] if isinstance(self[p], str)]


class Inflow(BaseSectionObject):
    """
    Section: [**INFLOWS**]

    Purpose:
        Specifies external hydrographs and pollutographs that enter the drainage system at specific nodes.

    Formats:
        ::

            Node FLOW   Tseries  (FLOW (1.0     Sfactor Base Pat))
            Node Pollut Tseries  (Type (Mfactor Sfactor Base Pat))

    Formats-PCSWMM:
        ``Node Parameter TimeSeries ParamType UnitsFactor ScaleFactor BaselineValue BaselinePattern``

    Formats-SWMM-GUI:
        ``Node Constituent TimeSeries Type Mfactor Sfactor Baseline Pattern``

    Remarks:
        External inflows are represented by both a constant and time varying component as follows:
        | Inflow = (Baseline value)*(Pattern factor) + (Scaling factor)*(Time series value)

        If an external inflow of a pollutant concentration is specified for a node,
        then there must also be an external inflow of FLOW provided for the same node, unless the node is an Outfall.
        In that case a pollutant can enter the system during periods
        when the outfall is submerged and reverse flow occurs.

    Examples:
        ::

            NODE2  FLOW N2FLOW
            NODE33 TSS  N33TSS CONCEN

            ;Mass inflow of BOD in time series N65BOD given in lbs/hr ;(126 converts lbs/hr to mg/sec)
            NODE65 BOD N65BOD MASS 126
            ;Flow inflow with baseline and scaling factor
            N176 FLOW FLOW_176 FLOW 1.0 0.5 12.7 FlowPat

    Args:
        Node (str): name of node where external inflow enters.
        Constituent (str): ``'FLOW'`` or name of pollutant. ``Pollut``
        TimeSeries (str): name of time series in [``TIMESERIES``] section describing how external flow or pollutant
            loading varies with time. ``Tseries``
        Type (str): ``'FLOW'`` or ``CONCEN`` if pollutant inflow is described as a concentration, ``MASS`` if it is
            described as a mass flow rate (default is ``CONCEN``).
        Mfactor (float): the factor that converts the inflow’s mass flow rate units into the project’s mass units per
            second, where the project’s mass units are those specified for the pollutant in the [POLLUTANTS] section (
            default is 1.0 - see example below).
        Sfactor (float): scaling factor that multiplies the recorded time series values (default is 1.0).
        Baseline (float): constant baseline value added to the time series value (default is 0.0). ``Base``
        Pattern (str): name of optional time pattern in [PATTERNS] section used to adjust the baseline value on a
            periodic basis. ``Pat``

    Attributes:
        Node (str): name of node where external inflow enters.
        Constituent (str): ``'FLOW'`` or name of pollutant. ``Pollut``
        TimeSeries (str): name of time series in [``TIMESERIES``] section describing how external flow or pollutant
            loading varies with time. ``Tseries``
        Type (str): ``'FLOW'`` or ``CONCEN`` if pollutant inflow is described as a concentration, ``MASS`` if it is
            described as a mass flow rate (default is ``CONCEN``).
        Mfactor (float): the factor that converts the inflow’s mass flow rate units into the project’s mass units per
            second, where the project’s mass units are those specified for the pollutant in the [POLLUTANTS] section (
            default is 1.0 - see example below).
        Sfactor (float): scaling factor that multiplies the recorded time series values (default is 1.0).
        Baseline (float): constant baseline value added to the time series value (default is 0.0). ``Base``
        Pattern (str): name of optional time pattern in [PATTERNS] section used to adjust the baseline value on a
            periodic basis. ``Pat``
    """
    _identifier = (IDENTIFIERS.Node, IDENTIFIERS.Constituent)
    _section_label = INFLOWS

    class TYPES:
        FLOW = 'FLOW'
        CONCEN = 'CONCEN'
        MASS = 'MASS'

    def __init__(self, Node, Constituent=TYPES.FLOW, TimeSeries=None, Type=TYPES.FLOW, Mfactor=1.0, Sfactor=1.0,
                 Baseline=0., Pattern=NaN):
        self.Node = str(Node)
        self.Constituent = str(Constituent)
        self.TimeSeries = TimeSeries
        self.Type = str(Type)
        self.Mfactor = float(Mfactor)
        self.Sfactor = float(Sfactor)
        self.Baseline = float(Baseline)
        self.Pattern = Pattern

        if (TimeSeries is None) or (TimeSeries == ''):
            self.TimeSeries = '""'


class Coordinate(BaseSectionObject):
    """
    Section: [**COORDINATES**]

    Purpose:
        Assigns X,Y coordinates to drainage system nodes.

    Format:
        ::

            Node Xcoord Ycoord

    Args:
        Node (str): name of node.
        x (float): horizontal coordinate relative to origin in lower left of map. ``Xcoord``
        y (float): vertical coordinate relative to origin in lower left of map. ``Ycoord``

    Attributes:
        Node (str): name of node.
        x (float): horizontal coordinate relative to origin in lower left of map. ``Xcoord``
        y (float): vertical coordinate relative to origin in lower left of map. ``Ycoord``
    """
    _identifier = IDENTIFIERS.Node
    _section_label = COORDINATES

    def __init__(self, Node, x, y):
        self.Node = str(Node)
        self.x = float(x)
        self.y = float(y)

    @property
    def point(self):
        return self.x, self.y

    def to_inp_line(self):
        # separate function to keep accuracy
        return f'{self.Node} {self.x:{GIS_FLOAT_FORMAT}} {self.y:{GIS_FLOAT_FORMAT}}'


class RainfallDependentInfiltrationInflow(BaseSectionObject):
    """
    Section: [**RDII**]

    Purpose:
        Specifies the parameters that describe rainfall-dependent infiltration/inflow (RDII)
        entering the drainage system at specific nodes.

    Format:
        ::

            Node UHgroup SewerArea

    Args:
        Node (str): name of node.
        hydrograph (str): name of an RDII unit hydrograph group specified in the [``HYDROGRAPHS``] section.
        sewer_area (float): area of the sewershed which contributes RDII to the node (acres or hectares).

    Attributes:
        Node (str): name of node.
        hydrograph (str): name of an RDII unit hydrograph group specified in the [``HYDROGRAPHS``] section.
        sewer_area (float): area of the sewershed which contributes RDII to the node (acres or hectares).
    """
    _identifier = IDENTIFIERS.Node
    _section_label = RDII

    def __init__(self, Node, hydrograph, sewer_area):
        self.Node = str(Node)
        self.hydrograph = str(hydrograph)
        self.sewer_area = float(sewer_area)


class Treatment(BaseSectionObject):
    """
    Section: [**TREATMENT**]

    Purpose:
        Specifies the degree of treatment received by pollutants at specific nodes of the drainage system.

    Format:
        ::

            Node Pollut Result = Func

    Args:
        Node (str): Name of node where treatment occurs.
        pollutant (str): Name of pollutant receiving treatment.
        result (str): Result computed by treatment function. Choices are C (function computes effluent concentration) R (function computes fractional removal).
        function (str): mathematical function expressing treatment result in terms of pollutant concentrations, pollutant removals, and other standard variables (see below).

    Attributes:
        Node (str): Name of node where treatment occurs.
        pollutant (str): Name of pollutant receiving treatment.
        result (str): Result computed by treatment function. Choices are C (function computes effluent concentration) R (function computes fractional removal).
        function (str): mathematical function expressing treatment result in terms of pollutant concentrations, pollutant removals, and other standard variables (see below).

    Remarks:
        Treatment functions can be any well-formed mathematical expression involving:
            - inlet pollutant concentrations (use the pollutant name to represent a concentration)
            - removal of other pollutants (use R\_ pre-pended to the pollutant name to represent removal)
            - process variables which include:
                - FLOW for flow rate into node (user’s flow units)
                - DEPTH for water depth above node invert (ft or m)
                - AREA for node surface area (ft2 or m2)
                - DT for routing time step (seconds)
                - HRT for hydraulic residence time (hours)

        Any of the following math functions can be used in a treatment function:
            - abs(x) for absolute value of x
            - sgn(x) which is +1 for x >= 0 or -1 otherwise
            - step(x) which is 0 for x <= 0 and 1 otherwise
            - sqrt(x) for the square root of x
            - log(x) for logarithm base e of x
            - log10(x) for logarithm base 10 of x
            - exp(x) for e raised to the x power
            - the standard trig functions (sin, cos, tan, and cot)
            - the inverse trig functions (asin, acos, atan, and acot)
            - the hyperbolic trig functions (sinh, cosh, tanh, and coth)

        along with the standard operators +, -, \*, /, ^ (for exponentiation ) and any level of nested parentheses.

    Examples:
        ::

            ; 1-st order decay of BOD
            Node23 BOD C = BOD * exp(-0.05*HRT)
            ; lead removal is 20% of TSS removal
            Node23 Lead R = 0.2 * R_TSS
    """
    _identifier = (IDENTIFIERS.Node, IDENTIFIERS.Pollutant)
    _section_label = TREATMENT

    def __init__(self, Node, pollutant, result, function):
        self.Node = str(Node)
        self.pollutant = str(pollutant)
        self.result = str(result)
        self.function = str(function)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        for name, pollutant, *line in multi_line_args:
            result, function = ' '.join(line).split('=')
            yield cls(name, pollutant, result, function)

    def to_inp_line(self):
        """
        convert object to one line of the ``.inp``-file

        for ``.inp``-file writing

        Returns:
            str: SWMM .inp file compatible string
        """
        return f'{self.Node} {self.pollutant} {self.result} = {self.function}'



from networkx import DiGraph, all_simple_paths, subgraph, node_connected_component

from ..inp import SwmmInput
from .collection import nodes_dict, links_dict
from .filter import create_sub_inp


def inp_to_graph(inp):
    """
    create a network of the model with the networkx package

    Args:
        inp (SwmmInput): inp-file data

    Returns:
        networkx.DiGraph: networkx graph of the model
    """
    # g = nx.Graph()
    g = DiGraph()
    for node in nodes_dict(inp).keys():
        g.add_node(node)
    for link in links_dict(inp).values():
        # if link.FromNode not in g:
        #     g.add_node(link.FromNode)
        # if link.ToNode not in g:
        #     g.add_node(link.ToNode)
        g.add_edge(link.FromNode, link.ToNode, label=link.Name)
    return g


def get_path(g, start, end):
    # dijkstra_path(g, start, end)
    return list(all_simple_paths(g, start, end))[0]


def get_path_subgraph(base, start, end):
    if isinstance(base, SwmmInput):
        g = inp_to_graph(base)
    else:
        g = base
    sub_list = get_path(g, start=start, end=end)
    sub_graph = subgraph(g, sub_list)
    return sub_list, sub_graph


def next_links(inp, node, g=None):
    if g is None:
        g = inp_to_graph(inp)
    links = links_dict(inp)
    for i in g.out_edges(node):
        yield links[g.get_edge_data(*i)['label']]


def next_links_labels(g, node):
    labels = []
    for i in g.out_edges(node):
        labels.append(g.get_edge_data(*i)['label'])
    return labels


def next_nodes(g, node):
    return list(g.successors(node))


def previous_links(inp, node, g=None):
    if g is None:
        g = inp_to_graph(inp)
    links = links_dict(inp)
    for i in g.in_edges(node):
        yield links[g.get_edge_data(*i)['label']]


def previous_links_labels(g, node):
    labels = []
    for i in g.in_edges(node):
        labels.append(g.get_edge_data(*i)['label'])
    return labels


def previous_nodes(g, node):
    return list(g.predecessors(node))


def links_connected(inp, node, g=None):
    if g is None:
        g = inp_to_graph(inp)
    links = links_dict(inp)
    next_ = []
    for i in g.out_edges(node):
        next_.append(links[g.get_edge_data(*i)['label']])
    previous_ = []
    for i in g.in_edges(node):
        previous_.append(links[g.get_edge_data(*i)['label']])
    return previous_, next_


def number_in_out(g, node):
    return len(list(g.predecessors(node))), len(list(g.successors(node)))


def downstream_nodes(graph, node):
    """
    get all nodes downstream of the node given

    only the direction of links defined in the INP file counts (not the elevation)

    Args:
        graph (networkx.DiGraph): network of the inp data
        node (str): node label

    Returns:
        list[str]: list of nodes downstream of the given node
    """
    return _downstream_nodes(graph,  node)


def _downstream_nodes(graph: DiGraph, node: str, node_list=None) -> list:
    if node_list is None:
        node_list = list()
    node_list.append(node)
    next_nodes = list(graph.successors(node))
    if next_nodes:
        for n in next_nodes:
            if n in node_list:
                continue
            node_list = _downstream_nodes(graph, n, node_list)
    return node_list


def upstream_nodes(graph, node):
    """
    get all nodes upstream of the node given

    only the direction of links defined in the INP file counts (not the elevation)

    Args:
        graph (networkx.DiGraph): network of the inp data
        node (str): node label

    Returns:
        list[str]: list of nodes upstream of the given node
    """
    return _upstream_nodes(graph,  node)


def _upstream_nodes(graph: DiGraph, node: str, node_list=None) -> list:
    if node_list is None:
        node_list = list()
    node_list.append(node)
    next_nodes = list(graph.predecessors(node))
    if next_nodes:
        for n in next_nodes:
            if n in node_list:
                continue
            node_list = _upstream_nodes(graph, n, node_list)
    return node_list


def get_network_forks(inp):
    # pd.DataFrame.from_dict(forks, orient='index')
    g = inp_to_graph(inp)
    nodes = nodes_dict(inp)
    forks = dict()
    for n in nodes:
        forks[n] = number_in_out(g, n)
    return forks


def split_network(inp, keep_node, split_at_node=None, keep_split_node=True, graph=None, init_print=True):
    """
    split model network at the ``split_at_node``-node and keep the part with the ``keep_node``-node.

    If you don't want to keep the ``split_at_node``-node toggle ``keep_split_node`` to False

    Set ``graph`` if a network-graph already exists (is faster).

    Notes:
        CONTROLS not supported

    Args:
        inp (SwmmInput): inp-file data
        keep_node (str): label of a node in the part you want to keep
        split_at_node (str | list | tuple | set): if you want to split the network,
                define the label of the node (or multiple) where you want to split it.
        keep_split_node (bool): if you want to keep the ``split_at_node`` node.
        graph (networkx.DiGraph): networkx graph of the model

    Returns:
        SwmmInput: filtered inp-file data
    """
    if graph is None:
        graph = inp_to_graph(inp)

    if split_at_node is not None:
        if isinstance(split_at_node, str):
            graph.remove_node(split_at_node)
        else:
            for n in split_at_node:
                graph.remove_node(n)

    if isinstance(graph, DiGraph):
        graph = graph.to_undirected()
    sub = subgraph(graph, node_connected_component(graph, keep_node))

    if init_print:
        print(f'Reduced Network from {len(graph.nodes)} nodes to {len(sub.nodes)} nodes.')

    # _______________
    final_nodes = list(sub.nodes)
    if split_at_node is not None and keep_split_node:
        if isinstance(split_at_node, str):
            final_nodes.append(split_at_node)
        else:
            final_nodes += list(split_at_node)
    final_nodes = set(final_nodes)

    # _______________
    return create_sub_inp(inp, final_nodes)


def conduit_iter_over_inp(inp, start, end):
    """
    iterate of the inp-file data

    only correct when FromNode and ToNode are in the correct direction
    doesn't look backwards if split node

    Args:
        inp (SwmmInput): inp-file data
        start (str): start node label
        end (str): end node label

    Yields:
        Conduit: input conduits
    """
    g = inp_to_graph(inp)

    if start and end:
        shortest_path_nodes = get_path(g, start, end)
        for n in shortest_path_nodes:
            for i in g.out_edges(n):
                if i[1] in shortest_path_nodes:
                    yield g.get_edge_data(*i)['label']
    # else:
    #     node = start
    #     while node:
    #         for l in next_links(g, node):
    #             yield l
    #
    #         # Todo: abzweigungen ...
    #         node = list(next_nodes(g, node))[0]

    # while True:
    #     found = False
    #     for i, c in inp[sec.CONDUITS].items():
    #         if c.FromNode == node:
    #             conduit = c
    #
    #             node = conduit.ToNode
    #             yield conduit
    #             found = True
    #             break
    #     if not found or (node is not None and (node == end)):
    #         break


import matplotlib.pyplot as plt
from matplotlib import patches

from .collection import nodes_dict, links_dict
from .geo import complete_vertices
from .graph import get_path_subgraph, links_connected
from ..section_labels import (MAP, COORDINATES, POLYGONS, SUBCATCHMENTS, VERTICES, JUNCTIONS, STORAGE, OUTFALLS,
                              XSECTIONS, )
from ..sections import Outfall, Polygon, SubCatchment


def set_inp_dimensions(inp, ax):
    map_dim = inp[MAP]['DIMENSIONS']
    x_min, x_max = map_dim['lower-left X'], map_dim['upper-right X']
    delta_x = x_max - x_min
    y_min, y_max = map_dim['lower-left Y'], map_dim['upper-right Y']
    delta_y = y_max - y_min
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)


def plot_map(inp, sc_connector=True, sc_center=True):  # TODO
    """

    Args:
        inp ():

    Returns:
        matplotlib.Figure, matplotlib.Axes:
    """
    fig, ax = plt.subplots()

    # for name, node in coords[::80].iterrows():
    #     ax.text(node.x, node.y, name, horizontalalignment='center', verticalalignment='baseline')

    ax.set_axis_off()
    ax.set_aspect(1.0)

    # ---------------------
    complete_vertices(inp)

    if VERTICES in inp:
        for link, vertices in inp[VERTICES].items():
            x, y = zip(*vertices.vertices)
            ax.plot(x, y, 'y-')

    # ---------------------
    if COORDINATES in inp:
        points = dict(inp[COORDINATES])
    else:
        points = dict()
    from shapely import geometry as shp
    points.update({poly.Subcatch: shp.Polygon(poly.polygon).centroid for poly in inp[POLYGONS].values()})

    if POLYGONS in inp:
        for poly in inp[POLYGONS].values():  # type: Polygon

            # ----------------
            # sub-catchment polygon
            ax.add_patch(patches.Polygon(poly.polygon, closed=True, fill=False, hatch='/'))

            # ----------------
            # center point of sub-catchment
            if sc_center:
                center = points[poly.Subcatch]
                ax.scatter(x=center.x, y=center.y, marker='s', c='k', zorder=999)

            # ----------------
            # center connector to sub-catchment
            if sc_connector:
                subcatch = inp[SUBCATCHMENTS][poly.Subcatch]  # type: SubCatchment
                outlet_point = points[subcatch.Outlet]
                center = points[poly.Subcatch]
                ax.plot([center.x, outlet_point.x], [center.y, outlet_point.y], 'r--')

    # ---------------------
    if COORDINATES in inp:
        coords = inp[COORDINATES].frame
        node_style = {
            JUNCTIONS: {'marker': '.', 'color': 'b'},
            STORAGE: {'marker': 's', 'color': 'g'},
            OUTFALLS: {'marker': '^', 'color': 'r'},

        }
        ax.scatter(x=coords.x, y=coords.y,
                   marker=node_style[JUNCTIONS]['marker'], c=node_style[JUNCTIONS]['color'],
                   edgecolors='k', zorder=999)

        for section in [STORAGE, OUTFALLS]:
            if section in inp:
                is_in_sec = coords.index.isin(inp[section].keys())
                ax.scatter(x=coords[is_in_sec].x, y=coords[is_in_sec].y,
                           marker=node_style[section]['marker'], c=node_style[section]['color'],
                           edgecolors='k', zorder=9999)

    # ---------------------
    fig.tight_layout()
    return fig, ax


class COLS:
    INVERT_ELEV = 'SOK'
    CROWN_ELEV = 'BUK'
    GROUND_ELEV = 'GOK'  # rim elevation
    STATION = 'x'
    WATER = 'water'
    # NODE_STATION = 'node'


def get_longitudinal_data(inp, start_node, end_node, out=None, zero_node=None):
    sub_list, sub_graph = get_path_subgraph(inp, start=start_node, end=end_node)

    if zero_node is None:
        zero_node = start_node

    keys = [COLS.STATION, COLS.INVERT_ELEV, COLS.CROWN_ELEV, COLS.GROUND_ELEV, COLS.WATER]

    res = {k: list() for k in keys}

    def _update_res(*args):
        for k, v in zip(keys, args):
            res[k].append(v)

    # ---------------
    nodes = nodes_dict(inp)
    # ---------------
    profile_height = 0
    # ---------------
    nodes_depth = None
    if out is not None:
        nodes_depth = out.get_part('node', sub_list, 'Depth_above_invert').mean().to_dict()
    # ---------------
    stations_ = list(iter_over_inp_(inp, sub_list, sub_graph))
    stations = dict(stations_)
    for node, x in stations_:
        n = nodes[node]
        sok = n.Elevation
        # ---------------
        gok = sok
        if isinstance(n, Outfall):
            gok += profile_height
        else:
            gok += n.MaxDepth
        # ---------------
        if out is not None:
            water = sok + nodes_depth[node]
        else:
            water = None
        # ------------------
        prior_conduit, following_conduit = links_connected(inp, node, sub_graph)

        if prior_conduit:
            prior_conduit = prior_conduit[0]
            profile_height = inp[XSECTIONS][prior_conduit.Name].Geom1
            sok_ = sok + prior_conduit.OutOffset
            buk = profile_height + sok_
            _update_res(x - stations[zero_node], sok_, buk, gok, water)

        if following_conduit:
            following_conduit = following_conduit[0]
            profile_height = inp[XSECTIONS][following_conduit.Name].Geom1
            sok_ = sok + following_conduit.InOffset
            buk = profile_height + sok_
            _update_res(x - stations[zero_node], sok_, buk, gok, water)

    return res


def get_water_level(inp, start_node, end_node, out, zero_node=None, absolute=True):
    nodes_depth = out.get_part('node', None, 'Depth_above_invert').mean().to_dict()
    nodes = nodes_dict(inp)
    x_list = list()
    water_level_list = list()
    stations_ = list(iter_over_inp(inp, start_node, end_node))
    stations = dict(stations_)
    sok = 0
    for node, x in stations_:
        x_list.append(x - stations.get(zero_node, 0))
        if absolute:
            sok = nodes[node].Elevation
        water_level_list.append(sok + nodes_depth[node])

    return {COLS.WATER: water_level_list, COLS.STATION: x_list}


def iter_over_inp_(inp, sub_list, sub_graph):
    links = links_dict(inp)

    x = 0
    for node in sub_list:
        yield node, x
        # ------------------
        out_edges = list(sub_graph.out_edges(node))
        if out_edges:
            following_conduit = links[sub_graph.get_edge_data(*out_edges[0])['label']]
            x += following_conduit.Length


def iter_over_inp(inp, start_node, end_node):
    sub_list, sub_graph = get_path_subgraph(inp, start=start_node, end=end_node)
    return iter_over_inp_(inp, sub_list, sub_graph)


def get_node_station(inp, start_node, end_node, zero_node=None):
    stations = dict(iter_over_inp(inp, start_node, end_node))
    if zero_node:
        return set_zero_node(stations, zero_node)
    return stations


def set_zero_node(stations, zero_node):
    return {node: stations[node] - stations[zero_node] for node in stations}


def plot_longitudinal(inp, start_node, end_node, out=None, ax=None, zero_node=None):
    res = get_longitudinal_data(inp, start_node, end_node, out, zero_node=zero_node)

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    ax.plot(res[COLS.STATION], res[COLS.INVERT_ELEV], c='k')
    ax.plot(res[COLS.STATION], res[COLS.GROUND_ELEV], c='brown', lw=0.5)
    ax.plot(res[COLS.STATION], res[COLS.CROWN_ELEV], c='k')
    bottom = ax.get_ylim()[0]

    # Ground Fill
    ax.fill_between(res[COLS.STATION], res[COLS.GROUND_ELEV], res[COLS.CROWN_ELEV], color='#C49B98', alpha=0.5)
    ax.fill_between(res[COLS.STATION], res[COLS.INVERT_ELEV], bottom, color='#C49B98', alpha=0.5)

    if out is not None:
        # Water line
        ax.plot(res[COLS.STATION], res[COLS.WATER], c='b', lw=0.7)
        # Water fill
        ax.fill_between(res[COLS.STATION], res[COLS.WATER], res[COLS.INVERT_ELEV], color='#00D7FF', alpha=0.7)
        # Conduit Fill
        ax.fill_between(res[COLS.STATION], res[COLS.CROWN_ELEV], res[COLS.WATER], color='#B0B0B0', alpha=0.5)
        ax.set_ylim(top=max([max(res[COLS.WATER]), ax.get_ylim()[1]]))
    else:
        # Conduit Fill
        ax.fill_between(res[COLS.STATION], res[COLS.CROWN_ELEV], res[COLS.INVERT_ELEV], color='#B0B0B0', alpha=0.5)

    ax.set_xlim(res[COLS.STATION][0], res[COLS.STATION][-1])
    ax.set_ylim(bottom=bottom)
    return fig, ax



import itertools

from ..section_labels import SUBCATCHMENTS
from .collection import links_dict, nodes_dict
from ..section_lists import NODE_SECTIONS, LINK_SECTIONS


def check_for_nodes(inp):
    """
    check if any link-end-node is missing

    Args:
        inp (swmm_api.SwmmInput): inp data

    Returns:
       tuple[set[swmm_api.input_file.sections.link._Link], set[str]]: set of corrupt links and a set of missing nodes
    """
    links = links_dict(inp)
    node_labels = set(nodes_dict(inp).keys())
    nodes_missing = set()
    links_corrupt = set()
    for link in links.values():
        if link.FromNode not in node_labels:
            nodes_missing.add(link.FromNode)
            links_corrupt.add(link)

        if link.ToNode not in node_labels:
            nodes_missing.add(link.ToNode)
            links_corrupt.add(link)

    return links_corrupt, nodes_missing


def check_for_duplicates(inp):
    """
    check if any node, link or subcatchment has a duplicate label in a different section

    Args:
        inp (swmm_api.SwmmInput): inp data

    Returns:
        tuple[set[str], set[str]]: set of duplicate nodes, links and subcatchments
    """

    nodes = {s: set(inp[s]) for s in NODE_SECTIONS + [SUBCATCHMENTS] if s in inp}

    nodes_duplicate = set()
    for s1, s2 in itertools.combinations(nodes.keys(), 2):
        nodes_duplicate |= nodes[s1] & nodes[s2]

    # -------------------
    links = {s: set(inp[s]) for s in LINK_SECTIONS if s in inp}

    links_duplicate = set()
    for s1, s2 in itertools.combinations(links.keys(), 2):
        links_duplicate |= links[s1] & links[s2]

    return nodes_duplicate, links_duplicate


def check_for_subcatchment_outlets(inp):
    """
    check if any subcatchments lost their outlets

    Args:
        inp (swmm_api.SwmmInput): inp data

    Returns:
        tuple[set[swmm_api.input_file.sections.subcatch.SubCatchment], set[str]]: : set of subcatchments_corrupt, outlets_missing
    """
    if SUBCATCHMENTS in inp:
        possible_outlets = set(nodes_dict(inp).keys()) | set(inp.SUBCATCHMENTS.keys())
        outlets = inp.SUBCATCHMENTS.frame.Outlet
        outlets_missing = set(outlets) - possible_outlets
        subcatchments_corrupt = {sc for sc in inp.SUBCATCHMENTS.values() if sc.Outlet in outlets_missing}
        return subcatchments_corrupt, outlets_missing
